@@ -1,6 +1,16 @@
-from ctypes import *
+from ctypes import (
+    c_char_p,
+    pointer,
+)
 import inspect
-from .lib import *
+from .lib import (
+    CStr,
+    CyFunc,
+    CyValue,
+    CyType,
+    CyResultCode,
+    CyLoadModuleFunc,
+)
 from . import lib
 from textwrap import dedent
 import builtins
@@ -27,41 +37,56 @@ def cstr(s) -> CStr:
     return CStr(c_char_p(s), len(s))
 
 
+def py_to_cyvalue(vm, value) -> CyValue:
+    match type(value):
+        case builtins.bool:
+            return lib.cyValueTrue() if value else lib.cyValueFalse()
+        case builtins.int:
+            return lib.cyValueInteger(value)
+        case builtins.float:
+            return lib.cyValueNumber(value)
+        case builtins.str:
+            return lib.cyValueGetOrAllocStringInfer(vm, cstr(value))
+        case _:
+            return lib.cyValueNone()
+
+
 def cyvalue_to_py(vm, cyvalue):
-    match CyType(cyValueGetTypeId(cyvalue).value):
+    match CyType(lib.cyValueGetTypeId(cyvalue).value):
         case CyType.CY_TypeNone:
             return None
         case CyType.CY_TypeBoolean:
-            return cyValueAsBool(cyvalue)
+            return lib.cyValueAsBool(cyvalue)
         case CyType.CY_TypeInteger:
-            return cyValueAsInteger(cyvalue)
+            return lib.cyValueAsInteger(cyvalue)
         case CyType.CY_TypeNumber:
-            return cyValueAsNumber(cyvalue)
+            return lib.cyValueAsNumber(cyvalue)
         case CyType.CY_TypeStaticAstring:
-            return cyValueToTempString(vm, cyvalue).charz.decode()
+            return lib.cyValueToTempString(vm, cyvalue).charz.decode()
         case CyType.CY_TypeStaticUstring:
-            return cyValueToTempString(vm, cyvalue).charz.decode()
+            return lib.cyValueToTempString(vm, cyvalue).charz.decode()
         case CyType.CY_TypeAstring:
-            return cyValueToTempString(vm, cyvalue).charz.decode()
+            return lib.cyValueToTempString(vm, cyvalue).charz.decode()
         case CyType.CY_TypeUstring:
-            return cyValueToTempString(vm, cyvalue).charz.decode()
+            return lib.cyValueToTempString(vm, cyvalue).charz.decode()
         case CyType.CY_TypeStringSlice:
-            return cyValueToTempString(vm, cyvalue).charz.decode()
+            return lib.cyValueToTempString(vm, cyvalue).charz.decode()
         case CyType.CY_TypeRawString:
-            return cyValueToTempRawString(vm, cyvalue).charz
+            return lib.cyValueToTempRawString(vm, cyvalue).charz
         case CyType.CY_TypeRawStringSlice:
-            return cyValueToTempRawString(vm, cyvalue).charz
+            return lib.cyValueToTempRawString(vm, cyvalue).charz
+        case CyType.CY_TypeList:
+            length = lib.cyListLen(cyvalue)
+            cylist = [lib.cyListGet(vm, cyvalue, i) for i in range(length)]
+            return [cyvalue_to_py(vm, cyval) for cyval in cylist]
+        case _:
+            return cyvalue
 
 
 def generate_callback_wrapper(func):
     sig = inspect.signature(func)
-    nargs = len(sig.parameters)
     return_type = sig.return_annotation
 
-    # TODO: probably not a robust way to do this
-    if 'self' in sig.parameters:
-        nargs = nargs - 1
-
     @CyFunc
     def wrapper(vm, args, nargs):
         # convert args based on registered function signature
@@ -74,18 +99,22 @@ def generate_callback_wrapper(func):
                 continue
             match param.annotation:
                 case builtins.str:
-                    _args.append(cyValueToTempString(vm, args[i]).charz.decode())
+                    _args.append(lib.cyValueToTempString(vm, args[i]).charz.decode())
                 case builtins.int:
-                    _args.append(int(cyValueAsNumber(args[i])))
+                    _args.append(int(lib.cyValueAsNumber(args[i])))
                 case builtins.float:
-                    _args.append(cyValueAsNumber(args[i]))
+                    _args.append(lib.cyValueAsNumber(args[i]))
                 case builtins.bool:
-                    _args.append(cyValueToBool(args[i]))
-                case lib.CyValue: # raw cyValue
+                    _args.append(lib.cyValueToBool(args[i]))
+                case lib.CyValue: # raw lib.cyValue
                     _args.append(args[i])
                 case _: # no type specified, try to auto-convert
                     _args.append(cyvalue_to_py(vm, args[i]))
             i += 1
+            # quit when all args have been consumed
+            # allows default parameters in wrapped functions
+            if i == nargs:
+                break
 
         raw_ret = func(*_args)
 
@@ -95,45 +124,34 @@ def generate_callback_wrapper(func):
         # convert return type
         # match return_type:
             # case None:
-            #     ret = cyValueNone()
+            #     ret = lib.cyValueNone()
             # case builtins.str:
-            #     ret = cyValueGetOrAllocStringInfer(vm, cstr(raw_ret))
+            #     ret = lib.cyValueGetOrAllocStringInfer(vm, cstr(raw_ret))
 
         return ret
 
-    return wrapper, nargs
+    return wrapper
 
 
-# TODO rename this something else, because the context manager syntax is gone
-class ContextModule:
-    def __init__(self, cyber, name) -> None:
+class Module:
+    def __init__(self, cyber, name, contents) -> None:
         self.cyber = cyber
         self.name = name
-        self.functions = []
-        self.variables = []
-
-    def function(self, name):
-        def _decorator(func):
-            # print(f'[ContextModule]: registering function: {name} {func}')
-            wrapper, nargs = generate_callback_wrapper(func)
-            self.functions.append((name, wrapper, nargs))
-            return func
-
-        return _decorator
-
-    # def variable(self, name, nargs=0):
-    #     def _decorator(func):
-    #         self.variables.append((name, CyFunc(func), nargs))
-
-    #     return _decorator
+        self.functions = contents['funcs']
+        self.variables = contents['vars']
+        self.build()
 
     def build(self):
         @CyLoadModuleFunc
         def load_module(vm, mod):
             # print(f'[ContextModule]: loading module: {self.name}')
-            for name, nargs, wrapper in self.functions:
+            for func_info in self.functions:
                 # print(f'[ContextModule]: registering function: {name}')
-                self.cyber.set_module_func(mod, name, nargs, wrapper)
+                self.cyber.set_module_func(mod, *func_info)
+
+            for var_info in self.variables:
+                self.cyber.set_module_var(mod, *var_info)
+
             return True
 
         self.load_module = load_module
@@ -146,19 +164,42 @@ class CyberVM:
         self.last_output_type = None
         self.last_output = None
 
-        self.vm = cyVmCreate()
+        self.vm = lib.cyVmCreate()
         self.modules = []
         self.pending_modules = {}
         self.pending_module_classes = []
 
     def add_module_loader(self, name, loader):
-        cyVmAddModuleLoader(self.vm, cstr(name), loader)
+        lib.cyVmAddModuleLoader(self.vm, cstr(name), loader)
 
     def set_module_func(self, mod, name, nargs, func):
-        cyVmSetModuleFunc(self.vm, mod, cstr(name), nargs, func)
+        lib.cyVmSetModuleFunc(self.vm, mod, cstr(name), nargs, func)
 
     def set_module_var(self, mod, name, value):
-        cyVmSetModuleVar(self.vm, mod, cstr(name), value)
+        lib.cyVmSetModuleVar(self.vm, mod, cstr(name), py_to_cyvalue(self.vm, value))
+
+    def _ensure_module(self, module_name):
+        if module_name not in self.pending_modules:
+            self.pending_modules[module_name] = {'funcs':[], 'vars':[]}
+    
+    def generate_wrappers(self, module_name, func_name, func):
+        sig = inspect.signature(func)
+        nargs = len(sig.parameters)
+        # TODO: probably not a robust way to do this
+        if 'self' in sig.parameters:
+            nargs = nargs - 1
+        sigs = [nargs]
+
+        optionals = 0
+        for name, param in sig.parameters.items():
+            if param.default != inspect._empty:
+                optionals += 1
+                sigs.append(nargs - optionals)
+
+        wrapper = generate_callback_wrapper(func)
+
+        for n in sigs:
+            self.pending_modules[module_name]['funcs'].append((func_name, n, wrapper))
 
     def function(self, name):
         if isinstance(name, str):
@@ -166,28 +207,28 @@ class CyberVM:
             func_name = name
             if '.' in name:
                 module_name, func_name = name.split('.')
+            self._ensure_module(module_name)
             def _decorator(func):
-                wrapper, nargs = generate_callback_wrapper(func)            
-
-                if module_name not in self.pending_modules:
-                    self.pending_modules[module_name] = {'funcs':[], 'vars':[]}
-                self.pending_modules[module_name]['funcs'].append((func_name, nargs, wrapper))
+                self.generate_wrappers(module_name, func_name, func)
                 return func
             return _decorator
         else:
             # we ARE the decorator
             func = name
-            
             module_name = 'core'
+            self._ensure_module(module_name)
             func_name = func.__name__
-
-            wrapper, nargs = generate_callback_wrapper(func)            
-
-            if module_name not in self.pending_modules:
-                self.pending_modules[module_name] = {'funcs':[], 'vars':[]}
-            self.pending_modules[module_name]['funcs'].append((func_name, nargs, wrapper))
-
+            self.generate_wrappers(module_name, func_name, func)
             return func
+        
+    def variable(self, name, value):
+        module_name = 'core'
+        var_name = name
+        if '.' in name:
+            module_name, var_name = name.split('.')
+
+        self._ensure_module(module_name)
+        self.pending_modules[module_name]['vars'].append((var_name, value))
 
     def module(self, name):
         if isinstance(name, str):
@@ -216,16 +257,20 @@ class CyberVM:
                 dec = self.function(f'{klass._name}.{func}')
                 dec(getattr(klass, func))
 
+        def get_vars(klass):
+            for var in [m for m in dir(klass) if not callable(getattr(klass, m)) and not m.startswith('__')]:
+                self.variable(f'{klass._name}.{var}', getattr(klass, var))
+
         for klass in self.pending_module_classes:
             get_funcs(klass)
+            get_vars(klass)
             for sub in klass.__subclasses__():
                 get_funcs(sub)
+                get_vars(sub)
 
         for module_name, contents in self.pending_modules.items():
-            mod = ContextModule(self, module_name)
+            mod = Module(self, module_name, contents)
             self.modules.append(mod)
-            mod.functions = contents['funcs']
-            mod.build()
 
         self.pending_modules.clear()
         self.pending_module_classes.clear()
@@ -235,9 +280,9 @@ class CyberVM:
 
         self.last_output_type = None
         self.last_output = None
-        result = cyVmValidate(self.vm, cstr(src))
+        result = lib.cyVmValidate(self.vm, cstr(src))
         self.last_result = CyResultCode(result)
-        return cyVmGetLastErrorReport(self.vm).charz
+        return lib.cyVmGetLastErrorReport(self.vm).charz
         # if self.last_result != CyResultCode.CY_Success:
         #     match self.last_result:
         #         case CyResultCode.CY_ErrorToken:
@@ -258,10 +303,10 @@ class CyberVM:
         self.last_output_type = None
         self.last_output = None
         out = CyValue()
-        result = cyVmEval(self.vm, cstr(src), pointer(out))
+        result = lib.cyVmEval(self.vm, cstr(src), pointer(out))
         self.last_result = CyResultCode(result)
         if self.last_result != CyResultCode.CY_Success:
-            report = cyVmGetLastErrorReport(self.vm).charz
+            report = lib.cyVmGetLastErrorReport(self.vm).charz
             match self.last_result:
                 case CyResultCode.CY_ErrorToken:
                     raise CyberTokenError(report)
@@ -275,7 +320,7 @@ class CyberVM:
                     raise CyberUnknownError(report)
             raise CyberUnknownError
         
-        self.last_output_type = CyType(cyValueGetTypeId(out).value)
+        self.last_output_type = CyType(lib.cyValueGetTypeId(out).value)
         self.last_output = cyvalue_to_py(self.vm, out)
         return self.last_output
 
